#include <thread>
#include <iostream>

#include "hpp/dds/dds.hpp"
#include "IDL/counter.hpp"

int main(int argc, char *argv[])
{
    //****************************** Domain participant ******************************************//
    const short int TYPE_CODE_MAX_SERIALIZED_LENGTH = 20000;
    const int32_t USER_DATA_MAX_LENGHT = 65536;

    // set factory qos
    dds::domain::qos::DomainParticipantFactoryQos factoryQos =
            dds::domain::DomainParticipant::participant_factory_qos();
    factoryQos << dds::core::policy::EntityFactory::ManuallyEnable();
    factoryQos->resource_limits.max_objects_per_thread(4096);
    dds::domain::DomainParticipant::participant_factory_qos(factoryQos);

    // set domain participant qos
    dds::domain::qos::DomainParticipantQos dpQos = dds::core::QosProvider::Default()
            .participant_qos(rti::core::builtin_profiles::qos_lib::baseline());

    rti::core::policy::DomainParticipantResourceLimits resource_limits_qos;
    resource_limits_qos.type_code_max_serialized_length(0);
    resource_limits_qos.reader_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.writer_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.type_object_max_serialized_length(TYPE_CODE_MAX_SERIALIZED_LENGTH);
    dpQos << resource_limits_qos;

    std::map<std::string,std::string> dpPropertyMap = {{"dds.transport.UDPv4.builtin.recv_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.send_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.parent.message_size_max", "65530"},
                                                       {"dds.transport.shmem.builtin.receive_buffer_size", "204857600"}};

    dpQos << rti::core::policy::Property(dpPropertyMap.begin(),dpPropertyMap.end());

    // Create a DomainParticipant
    dds::domain::DomainParticipant domainParticipant = dds::domain::DomainParticipant(1,dpQos);
    domainParticipant.enable();

    //******************************    create subscriber      ******************************************//

    // choose a topicName
    std::string topicName = "MY_IDL";

    // create topic with the domainParticipant, the topicName, the DataType
    dds::topic::Topic<Idl_Number> topic(domainParticipant,topicName);

    auto qosSub =  domainParticipant.default_subscriber_qos()
            << dds::core::policy::Partition("defaultPartition");

    // create subscriber with created domainParticipant
    dds::sub::Subscriber subscriber= dds::sub::Subscriber (domainParticipant, qosSub);

    //******************************    create data Reader      ******************************************//

    // create DataReaderQos object and set Reliability Qos and History Qos
    dds::sub::qos::DataReaderQos qosDr;
    std::map<std::string, std::string> qosDrPropertyMap = {{"dds.data_reader.history.memory_manager.fast_pool.pool_buffer_max_size", "3000000"},
                                                           {"reader_resource_limits.dynamically_allocate_fragmented_samples", "true"}};

    qosDr << rti::core::policy::Property(qosDrPropertyMap.begin(), qosDrPropertyMap.end());

    // configuring durability service
    dds::core::policy::Durability durabilityQoS;
    durabilityQoS.kind(dds::core::policy::DurabilityKind::PERSISTENT);

    durabilityQoS->direct_communication(true);

    qosDr << dds::core::policy::History::KeepAll()
          << dds::core::policy::ResourceLimits()
          << dds::core::policy::Reliability::Reliable()
          << durabilityQoS;

    // create DataReader opject with the subscriber, the topic, and the dataReaderQos
    dds::sub::DataReader<Idl_Number> dataReader(subscriber, topic, qosDr);


    //******************************    receive sample     ******************************************//

//    std::this_thread::sleep_for(std::chrono::seconds(10));
    // take any received data for 2 minutes
    if (dataReader.is_nil() == false)
    {
        auto start = std::chrono::high_resolution_clock::now();
        long difference = 0;

        while(difference < 120000)
        {
            // Taking data from data reader
            auto lnSamples = dataReader->take();
            if (lnSamples.length() == 0)
            {
                std::this_thread::sleep_for(std::chrono::milliseconds(10));
            }

            for (const auto& index : lnSamples)
            {
                if (index.info().valid())
                {
                    std::cerr << index.data().number() << std::endl;
                }
            }

            auto now = std::chrono::high_resolution_clock::now();
            difference = std::chrono::duration_cast<std::chrono::milliseconds>(now - start).count();
        }
    }
    std::this_thread::sleep_for(std::chrono::seconds(100));

}
